/*****************************************************************************
 * playlist.c : Playlist management functions
 *****************************************************************************
 * Copyright (C) 1999-2001 VideoLAN
 * $Id: playlist.c,v 1.13 2002/01/10 04:11:25 sam Exp $
 *
 * Authors: Samuel Hocevar <sam@zoy.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/
#include <stdlib.h>                                      /* free(), strtol() */
#include <stdio.h>                                              /* sprintf() */
#include <string.h>                                            /* strerror() */
#include <errno.h>                                                 /* ENOMEM */

#include <vlc/vlc.h>

#include "stream_control.h"
#include "input_ext-intf.h"

#include "playlist.h"

#define PLAYLIST_STOPPED 0
#define PLAYLIST_RUNNING 1

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static void RunThread ( playlist_t * );
static void NextItem  ( playlist_t * );

/*****************************************************************************
 * playlist_Create: create playlist
 *****************************************************************************
 * Create a playlist structure.
 *****************************************************************************/
playlist_t * playlist_Create ( vlc_object_t *p_parent )
{
    playlist_t *p_playlist;

    /* Allocate structure */
    p_playlist = vlc_object_create( p_parent, VLC_OBJECT_PLAYLIST );
    if( !p_playlist )
    {
        msg_Err( p_parent, "out of memory" );
        return NULL;
    }

    vlc_object_setparent( p_playlist, p_parent );

    p_playlist->p_input = NULL;
    p_playlist->i_status = PLAYLIST_RUNNING;
    p_playlist->i_index = -1;
    p_playlist->i_size = 0;
    p_playlist->pp_items = NULL;
    vlc_mutex_init( p_playlist, &p_playlist->change_lock );

    if( vlc_thread_create( p_playlist, &p_playlist->thread_id,
         "playlist", (vlc_thread_func_t)RunThread, (void *)p_playlist ) )
    {
        msg_Err( p_playlist, "cannot spawn playlist thread" );
        vlc_mutex_destroy( &p_playlist->change_lock );
        free( p_playlist );
        return NULL;
    }

    return p_playlist;
}

/*****************************************************************************
 * playlist_Destroy: destroy the playlist
 *****************************************************************************
 * Delete all items in the playlist and free the playlist structure.
 *****************************************************************************/
void playlist_Destroy( playlist_t * p_playlist )
{
    p_playlist->b_die = 1;

    vlc_thread_join( p_playlist, p_playlist->thread_id );
    vlc_mutex_destroy( &p_playlist->change_lock );

    free( p_playlist );
}

/*****************************************************************************
 * playlist_Add: add an item to the playlist
 *****************************************************************************
 * Add an item to the playlist at position i_pos. If i_pos is PLAYLIST_END,
 * add it at the end regardless of the playlist current size.
 *****************************************************************************/
int playlist_Add( vlc_object_t *p_this, int i_pos, const char * psz_item )
{
    playlist_t *p_playlist;

    p_playlist = vlc_object_find_playlist( p_this->p_vlc, FIND_CHILD );

    if( p_playlist == NULL )
    {
        msg_Dbg( p_this, "no playlist present, creating one" );
        p_playlist = playlist_Create( p_this->p_vlc->p_this );

        if( p_playlist == NULL )
        {
            return VLC_EGENERIC;
        }

        vlc_object_acquire( p_playlist );
    }

    msg_Warn( p_this, "adding playlist item  %s ", psz_item );

    vlc_mutex_lock( &p_playlist->change_lock );

    p_playlist->i_size++;
    p_playlist->pp_items = realloc( p_playlist->pp_items,
                                    p_playlist->i_size * sizeof(void*) );
    if( p_playlist->pp_items == NULL )
    {
        msg_Err( p_playlist, "out of memory" );
        vlc_mutex_unlock( &p_playlist->change_lock );
        vlc_object_release( p_playlist );
        return -1;
    }

    i_pos = p_playlist->i_size - 1;
    p_playlist->pp_items[i_pos] = malloc( sizeof( playlist_item_t ) );
    p_playlist->pp_items[i_pos]->psz_name = strdup( psz_item );
    p_playlist->pp_items[i_pos]->i_type = 0;
    p_playlist->pp_items[i_pos]->i_status = 0;

    vlc_mutex_unlock( &p_playlist->change_lock );
    vlc_object_release( p_playlist );

    return( 0 );
}

/*****************************************************************************
 * playlist_Next: switch to next playlist item
 *****************************************************************************
 * Switch to the next item of the playlist. If there is no next item, the
 * position of the resulting item is set to -1.
 *****************************************************************************/
void playlist_Next( playlist_t * p_playlist )
{
    vlc_mutex_lock( &p_playlist->change_lock );

    if( !p_playlist->i_size )
    {
        p_playlist->i_index = -1;
    }
    else
    {
        p_playlist->i_index++;
        if( p_playlist->i_index > p_playlist->i_size - 1 )
        {
            p_playlist->i_index = -1;
        }
    }

    vlc_mutex_unlock( &p_playlist->change_lock );
}

/*****************************************************************************
 * playlist_Prev: switch to previous playlist item
 *****************************************************************************
 * Switch to the previous item of the playlist. If there is no previous
 * item, the position of the resulting item is set to -1.
 *****************************************************************************/
void playlist_Prev( playlist_t * p_playlist )
{
}

/*****************************************************************************
 * playlist_Delete: delete an item from the playlist
 *****************************************************************************
 * Delete the item in the playlist with position i_pos.
 *****************************************************************************/
int playlist_Delete( playlist_t * p_playlist, int i_pos )
{
    return( 0 );
}

/*****************************************************************************
 * playlist_Jumpto: go to a specified position in playlist.
 *****************************************************************************/
void playlist_Jumpto( playlist_t * p_playlist , int i_pos)
{
}

/* URL-decode a file: URL path, return NULL if it's not what we expect */
void intf_UrlDecode( char *encoded_path )
{
    char *tmp = NULL, *cur = NULL, *ext = NULL;
    int realchar;

    if( !encoded_path || *encoded_path == '\0' )
    {
        return;
    }

    cur = encoded_path ;

    tmp = calloc(strlen(encoded_path) + 1,  sizeof(char) );

    while ( ( ext = strchr(cur, '%') ) != NULL)
    {
        strncat(tmp, cur, (ext - cur) / sizeof(char));
        ext++;

        if (!sscanf(ext, "%2x", &realchar))
        {
            free(tmp);
            return;
        }

        tmp[strlen(tmp)] = (char)realchar;

        cur = ext + 2;
    }

    strcat(tmp, cur);
    strcpy(encoded_path,tmp);
}

/* Following functions are local */

/*****************************************************************************
 * RunThread: main interface thread
 *****************************************************************************/
static void RunThread ( playlist_t *p_playlist )
{
    while( !p_playlist->b_die )
    {
        /* If there is an input, check that it doesn't need to die. */
        if( p_playlist->p_input )
        {
            if( p_playlist->p_input->i_status == THREAD_OVER )
            {
                input_thread_t *p_input;

                /* Unlink current input */
                vlc_mutex_lock( &p_playlist->change_lock );
                p_input = p_playlist->p_input;
                p_playlist->p_input = NULL;
                vlc_mutex_unlock( &p_playlist->change_lock );

                /* Destroy input */
                input_DestroyThread( p_input );
            }
            else if(    ( p_playlist->p_input->i_status == THREAD_READY
                           || p_playlist->p_input->i_status == THREAD_ERROR )
                     && ( p_playlist->p_input->b_error
                           || p_playlist->p_input->b_eof ) )
            {
                input_StopThread( p_playlist->p_input, NULL );
            }
        }
        else if( p_playlist->i_status != PLAYLIST_STOPPED )
        {
            /* Select the next playlist item */
            playlist_Next( p_playlist );

            /* don't loop by default: stop at playlist end */
            if( p_playlist->i_index == -1 )
            {
                p_playlist->i_status = PLAYLIST_STOPPED;
            }
            else
            {
                input_thread_t *p_input;

                //p_playlist->i_mode = PLAYLIST_FORWARD +
                //    config_GetInt( p_playlist, "loop-playlist" );
                msg_Dbg( p_playlist, "creating new input thread" );
                p_input = input_CreateThread( p_playlist->p_this,
                            p_playlist->pp_items[p_playlist->i_index], NULL );
                if( p_input != NULL )
                {
                    /* Link current input */
                    vlc_mutex_lock( &p_playlist->change_lock );
                    p_playlist->p_input = p_input;
                    vlc_mutex_unlock( &p_playlist->change_lock );
                }
            }
        }

        msleep( INTF_IDLE_SLEEP );
    }

    /* If there is an input, make him die */
    while( p_playlist->p_input )
    {
        if( p_playlist->p_input->i_status == THREAD_OVER )
        {
            input_thread_t *p_input;

            /* Unlink current input */
            vlc_mutex_lock( &p_playlist->change_lock );
            p_input = p_playlist->p_input;
            p_playlist->p_input = NULL;
            vlc_mutex_unlock( &p_playlist->change_lock );

            /* Destroy input */
            input_DestroyThread( p_input );
        }
        else if(    ( p_playlist->p_input->i_status == THREAD_READY
                       || p_playlist->p_input->i_status == THREAD_ERROR )
                 && ( p_playlist->p_input->b_error
                       || p_playlist->p_input->b_eof ) )
        {
            input_StopThread( p_playlist->p_input, NULL );
        }
        else
        {
            p_playlist->p_input->b_eof = 1;
        }

        msleep( INTF_IDLE_SLEEP );
    }
}

/*****************************************************************************
 * NextItem: select next playlist item
 *****************************************************************************
 * This function copies the next playlist item to the current structure,
 * depending on the playlist browsing mode.
 *****************************************************************************/
static void NextItem( playlist_t * p_playlist )
{
}

